# PIG Pilot (Update handlers)
# Copyright (C) 2005  Robin J Patenall
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#

# Basic system imports
import sys, time, random, math

# Pygame imports
import pygame

# PyOpenGL imports
try:
    from OpenGL.GL import *
    from OpenGL.GLU import *
except:
    print 'This entry requires PyOpenGL'
    raise SystemExit

# Texture handler import
import texture

# Sound handler import
import sound

# Text manager
import texfont


font = texfont.Font('res/AvantGarde-Demi.txf')

# The generic update handler
class Default :
  
  # The list of pygame events we are interested in
  events_wanted = []
  
  # The wanted frame delay
  frame_delay = 20
  
  # Perform an update.
  def update(self, screen_size, delta_time, events) :
    
    # In this basic handler just clear the screen.
    glClearColor(0.0, 1.0, 0.0, 0.0)
    glClear(GL_COLOR_BUFFER_BIT)
    
    # Stay in this handler.
    return None

# Test update handler that will wait for a escape to be pressed
class Wait_For_ESC(Default) :
   
  # The list of pygame events we are interested in
  events_wanted = [pygame.KEYDOWN]
  
  # The wanted frame delay
  frame_delay = 20
  
  # Perform an update.
  def update(self, screen_size, delta_time, events) :
    
    # The next handler to use, default to ourselves.
    handler_next = None
    
    # In this basic handler just clear the screen.
    glClearColor(1.0, 0.0, 0.0, 0.0)
    glClear(GL_COLOR_BUFFER_BIT)
    
    # If escape has been pressed ...
    for event in events :
      if event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE :
        
        # ... then skip to the ending handler.
        handler_next = Ender()
        
    # Return the appropriate handler.
    return handler_next

      
# Test update handler that will end the game
class Ender(Default) :
   
  # The list of pygame events we are interested in
  events_wanted = [pygame.KEYDOWN]
  
  # The wanted frame delay
  frame_delay = 20
  
  def __init__(self) :
    
    # Set the initial state.
    self.state = 0
    
    # The current offset.
    self.offset = 384.0
    
    # The maximum offset
    self.offset_max = 384.0
    
    # The move speed of the entries
    self.speed = 192.0
    
    # Try to load the text.
    self.text = [ \
      'PIG Pilot',
      'by',
      '(C) 2005 Robin J. Patenall',
      'AKA',
      'Robot Guy',
      'An entry in the',
      'PyWeek Game Programming Challenge'
    ]
  
  
  # Perform an update.
  def update(self, screen_size, delta_time, events) :

    # If we are in state 0 (raising)
    if self.state == 0 :
      
      # Set the frame delay
      self.frame_delay = 20
      
      # Decrease the offset.
      self.offset -= self.speed * delta_time
      
      # If we reached the top ...
      if self.offset < 0.0 :
        
        # ... fix it to the top and change states.
        self.offset = 0.0
        self.state = 1
        self.time_left = 4.0
      
    # If we are in state 1 (waiting)
    elif self.state == 1 :
      
      # Set the frame delay for waiting
      self.frame_delay = 200
      self.time_left -= delta_time
      
      # If we've timed out
      if self.time_left < 0 :
        self.state = 2
        self.frame_delay = 20        
          
    # If we are in state 2 (lowering)
    elif self.state == 2 :
      
      # Set the frame delay
      self.frame_delay = 20
      
      # Increase the offset.
      self.offset += self.speed * delta_time
      
      # If we reached the bottom ...
      if self.offset > self.offset_max :
        
        # ... fix it to the bottom and change states.
        self.offset = self.offset_max
        self.state = 0
        
        # Exit
        sys.exit()        
        
    # If we are waiting for a keypress)
    if self.state == 0 or self.state == 1 :
      
      # If the up koy has been pressed ...
      for event in events :
        if event.type == pygame.KEYDOWN :
          if event.key == pygame.K_RETURN or event.key == pygame.K_DOWN :
            sound.play('res/menu_clap.wav')
            self.state = 2
            self.frame_delay = 20
    
    # Clear the screen.
    glClearColor(0.0, 0.0, 0.0, 0.0)
    glClear(GL_COLOR_BUFFER_BIT)
    
    # Set OpenGL for 2D drawing
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glOrtho(0, 640, 0, 480, -0.5, 0.5)
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    # Draw the title
    draw_title()
    glTranslatef(64.0, 8.0 - self.offset, 0.0)
    #glColor4f(1.0, 1.0, 1.0, 0.7)
    glColor4f(0.0, 0.0, 0.0, 1.0)
    glRectf(8.0, 0.0, 496.0, 344.0)
    
    glAlphaFunc(GL_GEQUAL, 0.0625);
    glEnable(GL_ALPHA_TEST);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)    
    glEnable(GL_TEXTURE_2D);
    
    font.bind_texture()
    texture.clear_bind()
    
    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)    
    glColor3f(0.0, 1.0, 0.0)
    glTranslatef(8.0, 344.0, 0.0)
    for line in self.text :
      glTranslatef(0.0, -font.max_ascent, 0.0)
      glPushMatrix()
      left_side = (496.0 - font.get_width(line))/2
      glTranslatef(left_side, 0.0, 0.0)
      font.render(line)
      glPopMatrix()

    glDisable(GL_ALPHA_TEST);
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    
    # Return the new state
    return None
    

class Game(Default) :
  
  # The list of pygame events we are interested in
  events_wanted = [pygame.KEYDOWN]
  
  # The wanted frame delay
  frame_delay = 20
  
  # The rotation.
  rotation = 0
  
  # The rotation.
  translation = 0
  
  # The player position
  player_x = 0
  player_y = 0
  
  # The player velocity
  player_vx = 0
  player_vy = 0
  
  # Player's shots
  #player_shots = [[0.0, 1.0, 0.0, 0.2]]
  player_shots = []
  player_shot_timer = 0.0
  player_shot_delay = 0.25
  
  aliens = []
  alien_timer = 0.0
  alien_delay = 1.0
  
  alien_shots=[]
  
  player_score = 0
  
  player_shield = 100
  player_dead = False
  wait_timer = 0
  skill = 0.0
  
  # Perform an update.
  def update(self, screen_size, delta_time, events) :
  
    keys = pygame.key.get_pressed()
    if not self.player_dead :
      if keys[pygame.K_LEFT] and not keys[pygame.K_RIGHT] :
        self.player_vx -= 360.0 * delta_time
        if(self.player_vx < -90.0) : self.player_vx = -90.0
      elif not keys[pygame.K_LEFT] and keys[pygame.K_RIGHT] :
        self.player_vx += 360.0 * delta_time
        if(self.player_vx > +90.0) : self.player_vx = +90.0
      else :
        if self.player_vx > 0 :
          self.player_vx -= 30.0 * delta_time
          if(self.player_vx < 0.0) : self.player_vx = 0.0
        else :
          self.player_vx += 30.0 * delta_time
          if(self.player_vx > 0.0) : self.player_vx = 0.0
        
      if keys[pygame.K_UP] and not keys[pygame.K_DOWN] :
        self.player_vy += 1.0 * delta_time
        if(self.player_vy > +1.0) : self.player_vy = +1.0
      elif not keys[pygame.K_UP] and keys[pygame.K_DOWN] :
        self.player_vy -= 1.0 * delta_time
        if(self.player_vy < -1.0) : self.player_vy = -1.0
      else :
        if self.player_vy > -0.1 :
          self.player_vy -= 0.2 * delta_time
          if(self.player_vy < -0.1) : self.player_vy = -0.1
        else :
          self.player_vy += 0.2 * delta_time
          if(self.player_vy > -0.1) : self.player_vy = -0.1
    
      self.player_shot_timer -= delta_time
      if self.player_shot_timer < 0.0 :
        self.player_shot_timer = 0.0
      
      if keys[pygame.K_SPACE] and self.player_shot_timer == 0.0 and len(self.player_shots) < 50 :
        self.player_shots.append([self.player_x, self.player_y, 0.0, 2.0, 1])
        self.player_shot_timer = self.player_shot_delay
        sound.play('res/player_shot.wav')      
    
    self.player_x += self.player_vx * delta_time
    if(self.player_x > 360.0) : self.player_x -= 360.0
    if(self.player_x < -360.0) : self.player_x += 360.0
    self.player_y += self.player_vy * delta_time
    if(self.player_y > 1.0) : self.player_y = 1.0
    if(self.player_y < 0.0) : self.player_y = 0.0
    
    self.skill += delta_time / 100.0
    self.alien_timer -= delta_time
    while self.alien_timer < 0.0 :
      delay = ((1.0 / (1.0 + (self.skill / 3.0))) + 0.25)

      self.alien_timer += delay
      pr1 = 1.0 - (1.0 / (1.0 + (self.skill / 2.0)))
      pr2 = 1.0 - (1.0 / (1.0 + (self.skill / 4.0)))
      r = random.random() * (1.0 + pr1 + pr2)
      if(r < 1.0) :
        self.aliens.append([random.randint(0, 359), 15.0, 0, -2.0, 0, 1.0, 0.0, 0.33, 10])
      elif(r < 1.0 + pr1) :
        self.aliens.append([random.randint(0, 359), 15.0, 0, -2.0, 1, 2.0, 0.0, 0.5, 25])
      else :
        self.aliens.append([random.randint(0, 359), 15.0, 0, -2.0, 2, 3.0, 0.0, 0.66, 50])
            
    for shot in self.player_shots :
      shot[0] += shot[2] * delta_time
      if(shot[0] < -360.0) : shot[0] += 360.0
      if(shot[0] > 360.0) : shot[0] -= 360.0
      shot[1] += shot[3] * delta_time
      for alien in self.aliens :
        dx = alien[0] - shot[0]
        if(dx <= -180.0) : dx += 360.0
        if(dx > 180.0) : dx -= 360.0
        dy = alien[1] - shot[1]
        if abs(dy) < 0.1 :
          if abs(dx) < 8  :
            alien[5] -= 1.0
            shot[4] = 0
            if(alien[5] <= 0.0) :
              self.player_score += alien[8]
              sound.play('res/alien_die.wav')
            else :
              sound.play('res/alien_hit.wav')
    self.aliens = filter(lambda alien:alien[5] > 0.0, self.aliens)      
      
      
    self.player_shots = filter(lambda shot:shot[1] < 15.0 and shot[4] != 0, self.player_shots)
    for alien in self.aliens :
      alien[0] += alien[2] * delta_time
      if(alien[0] < -360.0) : alien[0] += 360.0
      if(alien[0] > 360.0) : alien[0] -= 360.0
      alien[1] += alien[3] * delta_time
      alien[6] += alien[7] * delta_time
      if(alien[6] >= 1.0) : 
        alien[6] = 0.0
        dx = (self.player_x - alien[0])/60.0
        if(dx < -3) : dx += 6
        if(dx > 3) : dx -= 6
        dy = self.player_y - alien[1]
        h = math.hypot(dx,dy) / 4.0
        
        dx2 = dx + (self.player_vx / 60.0)* h
        h = math.hypot(dx2,dy) / 4.0
        
        dx = dx + (self.player_vx / 60.0)* h
        h = math.hypot(dx,dy) / 4.0
        dx = 60.0  * dx / h
        dy = dy / h
        
        self.alien_shots.append([alien[0], alien[1], dx, dy, 1, 5])
    self.aliens = filter(lambda alien:alien[1] > 0.0, self.aliens)
    
    for shot in self.alien_shots :
      shot[0] += shot[2] * delta_time
      if(shot[0] < -360.0) : shot[0] += 360.0
      if(shot[0] > 360.0) : shot[0] -= 360.0
      y = shot[1]
      shot[1] += shot[3] * delta_time
      newy = shot[1] 
      dx = self.player_x - shot[0]
      if(dx < -360.0) : dx += 360.0
      if(dx > 360.0) : dx -= 360.0
      dy = self.player_y - shot[1]
      if abs(dy) < 0.1 :
      #if y >= self.player_y and newy <= self.player_y :
        if abs(dx) < 8  :
          if not self.player_dead :
            self.player_shield -= shot[5]

            if self.player_shield <= 0 :
              self.player_shield = 0
              self.player_dead = True
              self.wait_timer = 3.0
              sound.play('res/player_explode.wav')
            else :
              sound.play('res/player_bang.wav')
    self.alien_shots = filter(lambda shot:shot[1] > 0.0 and shot[4] != 0, self.alien_shots)   
    self.aliens = filter(lambda alien:alien[5] > 0.0, self.aliens)      

    if not self.player_dead :    
      self.player_shield += 1.0 * delta_time
      if self.player_shield > 100.0 : self.player_shield = 100.0
    else :
      self.wait_timer -= delta_time;
      if self.wait_timer < 0 :
        handler_next = Scores()
        handler_next.add_score(self.player_score)
        return handler_next
      
    #print(" " + str(self.player_x) + " " + str(self.player_vx) + "\n")
    # Set up the projection.
    screen_width, screen_height = screen_size
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    gluPerspective(45.0, float(screen_width) / float(screen_height), 0.5, 100.0);

    glTranslatef(0.0, 0.5, 0.0)
    #glRotatef(self.rotation, 0.0, 0.0, 1.0)
    
    #glMatrixMode(GL_TEXTURE)
    #glLoadIdentity()
    # Update the game.
    self.rotation += delta_time * 0.2
    if(self.rotation > 1.0) : self.rotation -= 1.0
    self.translation += delta_time
    if(self.translation > 1.0) : self.translation -= 1.0
    
    #glTranslatef(self.rotation, self.translation, 0.0)
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glRotatef(self.player_x, 0.0, 0.0, -1.0)
    # In this basic handler just clear the screen.
    #glClearColor(0.0, 0.0, 0.0, 0.0)
    #glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    glClear(GL_DEPTH_BUFFER_BIT)
    
    glEnable(GL_DEPTH_TEST)
    glDepthFunc(GL_LESS)
    glEnable(GL_FOG)
    glFogi(GL_FOG_MODE, GL_LINEAR)
    glFogfv(GL_FOG_COLOR, (1.0, 1.0, 1.0))
    glFogf(GL_FOG_START, 10.0)
    glFogf(GL_FOG_END, 15.0)
    glHint(GL_FOG_HINT, GL_DONT_CARE)
    
    # Bind the title texture.
    glEnable(GL_TEXTURE_2D)
    # Bind the title texture.
    texture.bind('res/wormhole.png')
    
    # Set the texture parameters
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST)
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
    
    depth = 40.0    
    glBegin(GL_QUADS)
    glTexCoord2f(0.0, 0.0)
    glVertex3f(-1.5, -1.5, -depth)
    glTexCoord2f(1.0, 0.0)
    glVertex3f(+1.5, -1.5, -depth)
    glTexCoord2f(1.0, 1.0)
    glVertex3f(+1.5, +1.5, -depth)
    glTexCoord2f(0.0, 1.0)
    glVertex3f(-1.5, +1.5, -depth)
    glEnd()
    glBegin(GL_TRIANGLE_STRIP)
    ang = 0.0
    ang_step = math.pi / 32.0
    ang_max = 2.0*math.pi - (ang_step / 2.0)
    tex = 0.0
    tex_step = (6.0 * ang_step) / (2.0 * math.pi)
    while ang < ang_max :
      y = -math.cos(ang)
      x = math.sin(ang) 
      glTexCoord2f(tex+self.rotation, self.translation)
      glVertex3f(x, y, -0.5)
      glTexCoord2f(tex+self.rotation, self.translation+depth)
      glVertex3f(x, y, -0.5-depth)
      ang += ang_step
      tex += tex_step
    glTexCoord2f(tex+self.rotation, self.translation)
    glVertex3f(0.0, -1.0, -0.5)
    glTexCoord2f(tex+self.rotation, self.translation+depth)
    glVertex3f(0.0, -1.0, -0.5-depth)
    glEnd()
        
    glMatrixMode(GL_TEXTURE)
    glLoadIdentity()
    
    glMatrixMode(GL_MODELVIEW);
    #glLoadIdentity()    
     # Bind the title texture.
    texture.bind('res/shot_1.png')
    
    # Set the texture parameters
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST)
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
    glAlphaFunc(GL_GEQUAL, 0.0625)
    glEnable(GL_ALPHA_TEST)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
    glEnable(GL_BLEND)
    for shot in self.player_shots :
      glPushMatrix()
      glRotatef(-shot[0], 0.0, 0.0, -1.0)
      glTranslatef(0.0, 0.0, -1.2-shot[1])
      glBegin(GL_QUADS)
      glTexCoord2f(0.0, 0.0)
      glVertex3f(-0.05, -1.0, 0.0)
      glTexCoord2f(1.0, 0.0)
      glVertex3f(+0.05, -1.0, 0.0)
      glTexCoord2f(1.0, 1.0)
      glVertex3f(+0.05, -0.9, 0.0)
      glTexCoord2f(0.0, 1.0)
      glVertex3f(-0.05, -0.9, 0.0)
      glEnd()
      glPopMatrix()
    
    texture.bind('res/shot_2.png')
    
    # Set the texture parameters
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST)
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
    glAlphaFunc(GL_GEQUAL, 0.0625)
    glEnable(GL_ALPHA_TEST)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
    glEnable(GL_BLEND)
    for shot in self.alien_shots :
      glPushMatrix()
      glRotatef(-shot[0], 0.0, 0.0, -1.0)
      glTranslatef(0.0, 0.0, -1.2-shot[1])
      glBegin(GL_QUADS)
      glTexCoord2f(0.0, 0.0)
      glVertex3f(-0.05, -1.0, 0.0)
      glTexCoord2f(1.0, 0.0)
      glVertex3f(+0.05, -1.0, 0.0)
      glTexCoord2f(1.0, 1.0)
      glVertex3f(+0.05, -0.9, 0.0)
      glTexCoord2f(0.0, 1.0)
      glVertex3f(-0.05, -0.9, 0.0)
      glEnd()
      glPopMatrix()
      
    texture.bind('res/aliens.png')
    
    # Set the texture parameters
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST)
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
    glAlphaFunc(GL_GEQUAL, 0.0625)
    glEnable(GL_ALPHA_TEST)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
    glEnable(GL_BLEND)
    for alien in self.aliens :
      glPushMatrix()
      alientex = alien[4] * 0.25
      glRotatef(-alien[0], 0.0, 0.0, -1.0)
      glTranslatef(0.0, 0.0, -1.2-alien[1])
      glBegin(GL_QUADS)
      glTexCoord2f(alientex, 0.0)
      glVertex3f(-0.1, -1.0, 0.0)
      glTexCoord2f(alientex+0.25, 0.0)
      glVertex3f(+0.1, -1.0, 0.0)
      glTexCoord2f(alientex+0.25, 1.0)
      glVertex3f(+0.1, -0.8, 0.0)
      glTexCoord2f(alientex, 1.0)
      glVertex3f(-0.1, -0.8, 0.0)
      glEnd()
      glPopMatrix()
          
    # Bind the title texture.
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity()    

    if not self.player_dead :
       # Bind the title texture.
      texture.bind('res/player.png')
    
      # Set the texture parameters
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST)
      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
      glAlphaFunc(GL_GEQUAL, 0.0625)
      glEnable(GL_ALPHA_TEST)
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
      glEnable(GL_BLEND)
   
      glBegin(GL_QUADS)
      glTexCoord2f(0.0, 0.0)
      glVertex3f(-0.1, -1.0, -1.2-self.player_y)
      glTexCoord2f(1.0, 0.0)
      glVertex3f(+0.1, -1.0, -1.2-self.player_y)
      glTexCoord2f(1.0, 1.0)
      glVertex3f(+0.1, -0.8, -1.2-self.player_y)
      glTexCoord2f(0.0, 1.0)
      glVertex3f(-0.1, -0.8, -1.2-self.player_y)
      glEnd()
    
    glDisable(GL_DEPTH_TEST)
    glDisable(GL_FOG)
    # Set OpenGL for 2D drawing
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glOrtho(0, 640, 0, 480, -0.5, 0.5)
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)    
    
    font.bind_texture()
    texture.clear_bind()
    
    glColor3f(0.0, 1.0, 0.0)
    glTranslatef(0.0, 460.0-font.max_ascent, 0.0)
    line = "Shield : "
    linewidth = font.get_width(line) + 10
    font.render(line)
    glDisable(GL_TEXTURE_2D)
    ypos = (font.max_ascent - 10.0) / 2.0
    glRectf(linewidth, ypos+5, 620, ypos + 15)
    glColor3f(1.0, 0.0, 0.0)
    barwidth = 620.0 - linewidth
    glRectf(linewidth + barwidth * (self.player_shield / 100.0), ypos+5, 620, ypos+15)
    glEnable(GL_TEXTURE_2D)
    glColor3f(0.0, 1.0, 0.0)
    glTranslatef(0.0, -font.max_ascent, 0.0)    
    font.render("Score : " + str(self.player_score))
    #glTranslatef(320.0, 0.0 , 0.0)
    #font.render("Lives : 3");
    glLoadIdentity();
    if self.player_dead :
      texture.bind('res/game_over.png')
    
      # Set the texture parameters
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST)
      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
    
      glBegin(GL_QUADS)
      glTexCoord2f(0.0, 0.0)
      glVertex2f(160.0, 120.0)
      glTexCoord2f(1.0, 0.0)
      glVertex2f(480.0, 120.0)
      glTexCoord2f(1.0, 1.0)
      glVertex2f(480.0, 360.0)
      glTexCoord2f(0.0, 1.0)
      glVertex2f(160.0, 360.0)
      glEnd()

    # Bind the title texture.
    glDisable(GL_ALPHA_TEST)
    glDisable(GL_BLEND);   
    glDisable(GL_TEXTURE_2D)
    
    # The next handler to use, default to ourselves.
    handler_next = None
    
     # If escape has been pressed ...
    for event in events :
      if event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE :
        
        # ... then skip to the ending handler.
        handler_next = Scores()
    
    # Stay in this handler.
    return handler_next

    
# Helper function to display the title
def draw_title() :
  
  # Bind the title texture.
  texture.bind('res/title.png')
    
  # Set the texture parameters
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP)
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP)
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
    
  # Enable textures and alpha test
  glEnable(GL_TEXTURE_2D)
  glEnable(GL_ALPHA_TEST)
  glAlphaFunc(GL_GREATER, 0.5)
    
  # Set a default colour
  glColor3f(1.0, 1.0, 1.0)
    
  # Draw the menu entries
  glBegin(GL_QUADS)
  glTexCoord2f(0.0, 1.0)
  glVertex2f(0.0, 480.0)
  glTexCoord2f(0.0, 0.0)
  glVertex2f(0.0, 352.0)
  glTexCoord2f(1.0, 0.0)
  glVertex2f(640.0, 352.0)
  glTexCoord2f(1.0, 1.0)
  glVertex2f(640.0, 480.0)
  glEnd()
  
  glDisable(GL_TEXTURE_2D)
  glDisable(GL_ALPHA_TEST)
  
  
  
# Update to draw the menu
class Menu(Default) :
   
  # The list of pygame events we are interested in
  events_wanted = [pygame.KEYDOWN]
  
  # The wanted frame delay
  frame_delay = 20
  
  # Initialise the data structures
  def __init__(self) :
    
    # Create the data for the button positions
    self.buttons = [ \
      [ \
        pygame.Rect([ 64, 272, 128, 80]), # The position of the left section
        pygame.Rect([192, 272, 256, 80]), # The position of the middle section
        pygame.Rect([448, 272, 128, 80]), # The position of the right section
      ],
      [ \
        pygame.Rect([ 64, 184, 128, 80]), # The position of the left section
        pygame.Rect([192, 184, 256, 80]), # The position of the middle section
        pygame.Rect([448, 184, 128, 80]), # The position of the right section
      ],
      [ \
        pygame.Rect([ 64,  96, 128, 80]), # The position of the left section
        pygame.Rect([192,  96, 256, 80]), # The position of the middle section
        pygame.Rect([448,  96, 128, 80]), # The position of the right section
      ],
      [ \
        pygame.Rect([ 64,   8, 128, 80]), # The position of the left section
        pygame.Rect([192,   8, 256, 80]), # The position of the middle section
        pygame.Rect([448,   8, 128, 80]), # The position of the right section
      ]
    ]
    
    self.textures_centre = [ \
      [0.0, 0.671875, 0.5, 0.984375],     
      [0.0, 0.343750, 0.5, 0.656250],
      [0.0, 0.015625, 0.5, 0.328125],
      [0.5, 0.671875, 1.0, 0.984375],
    ]
    
    self.textures_sides = [\
      [ \
        [0.50, 0.343750, 0.75, 0.656250],
        [0.75, 0.343750, 1.00, 0.656250]
      ], 
      [ \
        [0.50, 0.015625, 0.75, 0.328125],
        [0.75, 0.015625, 1.00, 0.328125]
      ]
    ]
    
    # Set the button selected
    self.button_selected = 0
    
    # Set the initial state.
    self.state = 0
    
    # The current offset.
    self.offset = 384.0
    
    # The maximum offset
    self.offset_max = 384.0
    
    # The move speed of the entries
    self.speed = 192.0
    
    # The button that was pressed.
    self.button_pressed = -1
    
    
  # Perform an update.
  def update(self, screen_size, delta_time, events) :

    # Helper function to draw a textured quad.
    def draw_quad(screen_rect, texture_rect) :
            
      # Draw the quad
      glTexCoord2f(texture_rect[0], texture_rect[1])
      glVertex2fv(screen_rect.topleft)
      glTexCoord2f(texture_rect[0], texture_rect[3])
      glVertex2fv(screen_rect.bottomleft)
      glTexCoord2f(texture_rect[2], texture_rect[3])
      glVertex2fv(screen_rect.bottomright)
      glTexCoord2f(texture_rect[2], texture_rect[1])
      glVertex2fv(screen_rect.topright)
    
    
    # Update out state.
    update_handler_new = None
    
    # If we are in state 0 (raising)
    if self.state == 0 :
      
      # Set the frame delay
      self.frame_delay = 20
      
      # Decrease the offset.
      self.offset -= self.speed * delta_time
      
      # If we reached the top ...
      if self.offset < 0.0 :
        
        # ... fix it to the top and change states.
        self.offset = 0.0
        self.state = 1
    
    # If we are in state 1 (waiting)
    elif self.state == 1 :
       
      # Set the frame delay for waiting
      self.frame_delay = -1
        
    # If we are in state 2 (lowering)
    elif self.state == 2 :
      
      # Set the frame delay
      self.frame_delay = 20
      
      # Increase the offset.
      self.offset += self.speed * delta_time
      
      # If we reached the bottom ...
      if self.offset > self.offset_max :
        
        # ... fix it to the bottom and change states.
        self.offset = self.offset_max
        self.state = 0

        # If the intro button was pressed ...
        if self.button_pressed == 0 :
          
          # ... switche to the intro update
          update_handler_new = Intro()
               
        # If the scores button was pressed ...
        elif self.button_pressed == 1 :
          
          # ... switche to the scores update
          update_handler_new = Game()
          #update_handler_new.add_score(random.randint(10, 20000))
        
        # If the scores button was pressed ...
        elif self.button_pressed == 2 :
          
          # ... switche to the scores update
          update_handler_new = Scores()
          #update_handler_new.add_score(random.randint(10, 20000))
      
        # If the quit button was pressed ...
        elif self.button_pressed == 3 :
          
          # ... switche to the ender update
          update_handler_new = Ender()
          
    # If we are waiting for a keypress)
    if self.state == 0 or self.state == 1 :
      
      # If the up koy has been pressed ...
      for event in events :
        if event.type == pygame.KEYDOWN :
          if event.key == pygame.K_UP :
            sound.play('res/menu_blip.wav')
            self.button_selected -= 1
            if self.button_selected < 0 :
              self.button_selected += len(self.buttons)
          elif event.key == pygame.K_DOWN :
            sound.play('res/menu_blip.wav')
            self.button_selected += 1
            if self.button_selected >= len(self.buttons) :
              self.button_selected -= len(self.buttons)
          elif event.key == pygame.K_RETURN :
            sound.play('res/menu_clap.wav')
            self.button_pressed = self.button_selected
            self.state = 2
            self.frame_delay = 20
  
    # Clear the screen.
    glClearColor(0.0, 0.0, 0.0, 0.0)
    glClear(GL_COLOR_BUFFER_BIT)
    
    # Set OpenGL for 2D drawing
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glOrtho(0, 640, 0, 480, -0.5, 0.5)
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    # Draw the title
    draw_title()
    
    # Bind the button texture.
    texture.bind('res/buttons.png')
    
    # Set the texture parameters
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
    
    # Enable textures and alpha test
    glEnable(GL_TEXTURE_2D)
    glEnable(GL_ALPHA_TEST)
    glAlphaFunc(GL_GREATER, 0.5)
    
    # Set a default colour
    glColor3f(1.0, 1.0, 1.0)
    
    # Draw the menu entries
    glBegin(GL_QUADS)
    for buttonid in range(4) :
      button = self.buttons[buttonid]
      if self.button_selected == buttonid :
        sides = self.textures_sides[1]
      else :
        sides = self.textures_sides[0]
      draw_quad(button[0].move(0, -self.offset), sides[0]);
      draw_quad(button[1].move(0, -self.offset), self.textures_centre[buttonid]);
      draw_quad(button[2].move(0, -self.offset), sides[1]);
    glEnd()

    glDisable(GL_TEXTURE_2D)
    glDisable(GL_ALPHA_TEST)
    
    # Return the new state
    return update_handler_new

# Update to draw the Introduction
class Intro(Default) :
   
  # The list of pygame events we are interested in
  events_wanted = [pygame.KEYDOWN]
  
  # The wanted frame delay
  frame_delay = 20
  
  def __init__(self) :
    
    # Set the initial state.
    self.state = 0
    
    # The current offset.
    self.offset = 384.0
    
    # The maximum offset
    self.offset_max = 384.0
    
    # The move speed of the entries
    self.speed = 192.0
    
    # Try to load the text.
    file = open('res/intro.txt')
    text = file.read()
    self.text = eval(text)
    self.page = 0
  
  
  # Perform an update.
  def update(self, screen_size, delta_time, events) :

    # Update out state.
    update_handler_new = None
    
    # If we are in state 0 (raising)
    if self.state == 0 :
      
      # Set the frame delay
      self.frame_delay = 20
      
      # Decrease the offset.
      self.offset -= self.speed * delta_time
      
      # If we reached the top ...
      if self.offset < 0.0 :
        
        # ... fix it to the top and change states.
        self.offset = 0.0
        self.state = 1
      
    # If we are in state 1 (waiting)
    elif self.state == 1 :
      
      # Set the frame delay for waiting
      self.frame_delay = -1
          
    # If we are in state 2 (lowering)
    elif self.state == 2 :
      
      # Set the frame delay
      self.frame_delay = 20
      
      # Increase the offset.
      self.offset += self.speed * delta_time
      
      # If we reached the bottom ...
      if self.offset > self.offset_max :
        
        # ... fix it to the bottom and change states.
        self.offset = self.offset_max
        self.state = 0
        self.page = 0
    
        update_handler_new = Menu()
        update_handler_new.button_selected = 1
        
        
    # If we are waiting for a keypress)
    if self.state == 0 or self.state == 1 :
      
      # If the up koy has been pressed ...
      for event in events :
        if event.type == pygame.KEYDOWN :
          if event.key == pygame.K_RETURN or event.key == pygame.K_DOWN :
            sound.play('res/menu_clap.wav')
            if self.page < (len(self.text) - 1) :
              self.page += 1
            else :
              self.state = 2
              self.frame_delay = 20
          elif event.key == pygame.K_UP :
            sound.play('res/menu_clap.wav')
            if self.page > 0 :
              self.page -= 1
          elif event.key == pygame.K_ESCAPE :
            sound.play('res/menu_clap.wav')
            self.state = 2
            self.frame_delay = 20
    
    # Clear the screen.
    glClearColor(0.0, 0.0, 0.0, 0.0)
    glClear(GL_COLOR_BUFFER_BIT)
    
    # Set OpenGL for 2D drawing
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glOrtho(0, 640, 0, 480, -0.5, 0.5)
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    

    # Draw the title
    draw_title()
    glTranslatef(64.0, 8.0 - self.offset, 0.0)
    #glColor4f(1.0, 1.0, 1.0, 0.7)
    #glColor4f(0.0, 0.0, 0.0, 1.0)
    #glRectf(8.0, 0.0, 496.0, 344.0)
    
    glAlphaFunc(GL_GEQUAL, 0.0625);
    glEnable(GL_ALPHA_TEST);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)    
    glEnable(GL_TEXTURE_2D);
    
    font.bind_texture()
    texture.clear_bind()
    
    glColor3f(0.0, 1.0, 0.0)
    glTranslatef(8.0, 344.0, 0.0)
    for line in self.text[self.page] :
      glTranslatef(0.0, -font.max_ascent, 0.0)
      glPushMatrix()
      font.render(line)
      glPopMatrix()

    glDisable(GL_ALPHA_TEST);
    glDisable(GL_BLEND);
    
    # Return the new state
    return update_handler_new
  
# Update to draw the Scored
class Scores(Default) :
   
  # The list of pygame events we are interested in
  events_wanted = [pygame.KEYDOWN]
  
  # The wanted frame delay
  frame_delay = 20
  
  def __init__(self) :
    
    # Set the initial state.
    self.state = 0
    
    # The current offset.
    self.offset = 384.0
    
    # The maximum offset
    self.offset_max = 384.0
    
    # The move speed of the entries
    self.speed = 192.0
    
    # The flash rate
    self.flash = True
    self.flash_time = 0.0
    self.flash_rate = 1.0
    
    # Flag if we are edititg.
    self.edit = None
    
    # Flag if we have changed
    self.changed = False
    
    # Try to load the text.
    try :
      file = open('res/scores.txt', 'r')
      scores = file.read()
      self.scores = eval(scores)
    except :
      self.scores = \
      [\
        ['Robot Guy',  2500, 0, 0], ['Robot Guy',  2250, 0, 0], 
        ['Robot Guy',  2000, 0, 0], ['Robot Guy',  1750, 1, 0],
        ['Robot Guy',  1500, 0, 0], ['Robot Guy',  1250, 0, 0], 
        ['Robot Guy',  1000, 0, 0], ['Robot Guy',   750, 0, 0],
        ['Robot Guy',   500, 0, 0], ['Robot Guy',   250, 0, 0], 
      ]
      self.changed = True
    self.sort_scores()
    
  
  # Sort the scores
  def sort_scores(self) :
    self.scores.sort(lambda a,b :-cmp(a[1], b[1]) or cmp(a[2], b[2]))
    
  # Save the scores
  def __del__(self) :
    
    # If the scores have changed ...
    if self.changed :
      # Try to save the data
      try :
        file = open('res/scores.txt', 'w')
        file.write(repr(self.scores))
        file.close
      except :
        pass
    
  # Add a score to the score system.
  def add_score(self, new_score) :
    
    # Ensure that the scores are sorted.
    self.sort_scores()
    
    # If the score is greater than the least score ...
    if self.scores[-1][1] < new_score :
      
      # ... we need to edit it.
      # Make sure that the markers are clear.
      for score in self.scores : score[3] = 0
      
      # Overwrite the smallest score with the new one.
      self.scores[-1][0] = ''
      self.scores[-1][1] = new_score
      self.scores[-1][2] = time.time()
      self.scores[-1][3] = 1
      
      # Resort the list ...
      self.sort_scores()
      
      # ... and mark us as changed
      self.changed = True
      
      # Find the new score
      self.edit = 0
      while self.edit < len(self.scores) :
        if self.scores[self.edit][3] == 1 : break
        self.edit += 1
  
  # Perform an update.
  def update(self, screen_size, delta_time, events) :

    # Update out state.
    update_handler_new = None
    
    # Update the flashing
    self.flash_time += delta_time
    while self.flash_time > self.flash_rate :
      self.flash_time -= self.flash_rate
      self.flash = not self.flash
      
   # If we are in state 0 (raising)
    if self.state == 0 :
      
      # Set the frame delay
      self.frame_delay = 20
      
      # Decrease the offset.
      self.offset -= self.speed * delta_time
      
      # If we reached the top ...
      if self.offset < 0.0 :
        
        # ... fix it to the top and change states.
        self.offset = 0.0
        self.state = 1
        
        # If we are going to be editing ...
        if self.edit != None :
          pygame.key.set_repeat(500, 30)
       
    # If we are in state 1 (waiting for a keypress)
    elif self.state == 1 :
      
      # If we are editing ...
      if self.edit != None :
        
        # Set the frame delay so we can flash
        self.frame_delay = 100
      else :
        
        # Otherwise, set the frame delay for waiting
        self.frame_delay = -1
   
    # If we are in state 2 (lowering)
    elif self.state == 2 :
      
      # Set the frame delay
      self.frame_delay = 20
      
      # Increase the offset.
      self.offset += self.speed * delta_time
      
      # If we reached the bottom ...
      if self.offset > self.offset_max :
        
        # ... fix it to the bottom and change states.
        self.offset = self.offset_max
        self.state = 0
        self.page = 0

        update_handler_new = Menu()
        update_handler_new.button_selected = 1
    
    # If we are editing ...
    if self.edit != None :
      
      # ... only respond to keypresses in the waiting state.
      if self.state == 1 :
        
        # The name being edited.
        score_name = self.scores[self.edit][0]
        
        # If the up koy has been pressed ...
        for event in events :
          if event.type == pygame.KEYDOWN :
            if event.key == pygame.K_RETURN :
              score_name = score_name.lstrip()
              if len(score_name) == 0 :
                score_name = 'Anonymous'
              self.scores[self.edit][0] = score_name
              sound.play('res/menu_clap.wav')
              self.edit = None
              pygame.key.set_repeat(0, 0)
            elif event.key == pygame.K_BACKSPACE :
              if len(score_name) > 0 :
                sound.play('res/menu_blip.wav')
                self.scores[self.edit][0] = score_name[:-1]
                self.flash = True
                self.flash_time = 0
            elif event.key >= 32 and event.key < 127 :
              if font.get_width(score_name) < 256 :
                sound.play('res/menu_blip.wav')
                key = chr(event.key)
                if (pygame.key.get_mods() & pygame.KMOD_SHIFT) != 0 :
                  key = key.upper()
                self.scores[self.edit][0] = score_name + key
                self.flash = True
                self.flash_time = 0
           
                
    # If we are not editing ...
    else :
      
      # ... we respond to key presses in both raising and waiting modes
      if self.state == 0 or self.state == 1 :
        
        # If the up koy has been pressed ...
        for event in events :
          if event.type == pygame.KEYDOWN :
            if event.key == pygame.K_RETURN or event.key == pygame.K_ESCAPE :
              sound.play('res/menu_clap.wav')
              self.state = 2
              self.frame_delay = 20
        
    # Clear the screen.
    glClearColor(0.0, 0.0, 0.0, 0.0)
    glClear(GL_COLOR_BUFFER_BIT)
    
    # Set OpenGL for 2D drawing
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glOrtho(0, 640, 0, 480, -0.5, 0.5)
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    # Draw the title
    draw_title()
    glTranslatef(64.0, 8.0 - self.offset, 0.0)
    #glColor4f(1.0, 1.0, 1.0, 0.7)
    #glColor4f(0.0, 0.0, 0.0, 1.0)
    #glRectf(8.0, 0.0, 496.0, 344.0)
    
    glAlphaFunc(GL_GEQUAL, 0.0625);
    glEnable(GL_ALPHA_TEST);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)    
    
    font.bind_texture()
    texture.clear_bind()
    
    glColor3f(0.0, 1.0, 0.0)
    glTranslatef(8.0, 344.0, 0.0)
    for scoreidx in range(len(self.scores)) :
      score = self.scores[scoreidx]
      glTranslatef(0.0, -font.max_ascent, 0.0)
      right_side = 496.0 - font.get_width(str(score[1]))
      glPushMatrix()
      if self.edit != None and scoreidx == self.edit and self.flash :
        font.render(score[0] + '_')
      else :
        font.render(score[0])        
      glPopMatrix()
      glPushMatrix()
      glTranslatef(right_side, 0.0, 0.0)
      font.render(str(score[1]))
      glPopMatrix()
    
    # If we are editing ...
    if self.edit != None:
      
      # ... skip a couple of lines ...
      glTranslatef(0.0, -font.max_ascent*2, 0.0)
      editstr = 'Enter your name'
      width = font.get_width(editstr)
      glTranslatef((496.0 - width) / 2, 0.0, 0.0)
      font.render(editstr)

    glDisable(GL_ALPHA_TEST);
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    
    # Return the new state
    return update_handler_new
      